#include <GL/glut.h> //glut library, it imports gl.h, glu.h
#include <stdlib.h> //for exit call
#include <iostream> //for FILE

/* easy indexing for texture array */
#define CEMENT 0

/* structure for the storing the image data */
struct Image 
{
    unsigned long size_x;
    unsigned long size_y;
    char *data;
};

typedef struct Image Image;

const int textureCount = 1; // specifies # of textures

Image myTextureData[textureCount];  //array storing image texture info
GLuint theTexture[textureCount];  //array storing OpenGL texture info

/* texture filename list */
char* textureFilenames[textureCount] = {"cement.bmp"};

int window_width = 800; //sets the width of the window
int window_height = 600; //sets the hight of the window

/*
=================
imageLoaderer
-loads our bitmap data into the Image struct
=================
*/
int imageLoader(const char *filename, Image *image) 
{
    FILE *file;

    unsigned long size;
    unsigned long i;
    unsigned short int planes;
    unsigned short int bpp;

    char temp;
	char finalName[80];

	glTexCoord2f(1.0, 0.0);

	strcpy(finalName, "" );
	strcat(finalName, filename);

    if ((file = fopen(finalName, "rb"))==NULL) 
	{
		printf("File Not Found : %s\n",finalName);
		return 0;
    }

    fseek(file, 18, SEEK_CUR);

	glTexCoord2f(1.0, 0.0);
    
	if ((i = fread(&image->size_x, 4, 1, file)) != 1) 
	{
		printf("Error reading width from %s.\n", finalName);
		return 0;
    }

    if ((i = fread(&image->size_y, 4, 1, file)) != 1) 
	{
		printf("Error reading height from %s.\n", finalName);
		return 0;
    }

    size = image->size_x * image->size_y * 3;

    if ((fread(&planes, 2, 1, file)) != 1) 
	{
		printf("Error reading planes from %s.\n", finalName);
		return 0;
    }

    if (planes != 1) 
	{
		printf("Planes from %s is not 1: %u\n", finalName, planes);
		return 0;
    }

    if ((i = fread(&bpp, 2, 1, file)) != 1) 
	{
		printf("Error reading bpp from %s.\n", finalName);
		return 0;
    }

    if (bpp != 24) 
	{
		printf("Bpp from %s is not 24: %u\n", finalName, bpp);
		return 0;
    }

    fseek(file, 24, SEEK_CUR);

    image->data = (char *) malloc(size);

    if (image->data == NULL) 
	{
		printf("Error allocating memory for color-corrected image data");
		return 0;
    }

    if ((i = fread(image->data, size, 1, file)) != 1) 
	{
		printf("Error reading image data from %s.\n", finalName);
		return 0;
    }

    for (i=0;i<size;i+=3) 
	{
		temp = image->data[i];
		image->data[i] = image->data[i+2];
		image->data[i+2] = temp;
    }
    return 1;
}

/*
=================
textureLoader
-physically loads the BMPs and converts them to texture data
=================
*/
void textureLoader() 
{
	/* set the pixel storage, GL_UNPACK_ALIGNMENT : specifies alignment requirements
	   for the start of each pixel row in memory 1 = byte-alignment*/
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	for(int k=0; k < textureCount; k++) 
	{
		if(!imageLoader(textureFilenames[k], &myTextureData[k])) 
			exit(1);
		
		/* generate texture names */
		glGenTextures(1, &theTexture[k]);
		/* create a named texture bound to a texture target */
	    glBindTexture(GL_TEXTURE_2D, theTexture[k]);

		/* set the texture parameters */
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_NEAREST);
		/* builds and load a set of mipmaps */
	   	gluBuild2DMipmaps(GL_TEXTURE_2D, 3, myTextureData[k].size_x, myTextureData[k].size_y, GL_RGB, GL_UNSIGNED_BYTE, myTextureData[k].data);
	}
}

/*
=================
boxCreator
-creates the sides of a 3D box
=================
*/
void boxCreator()
{
	/* bind a texture to a texture target */
	glBindTexture(GL_TEXTURE_2D, theTexture[CEMENT]);
	
	/* right-front */
	glBegin(GL_QUADS);
		/* set the texture coordinates for the given vertex*/
		glTexCoord2f(0.0,0.0);
		glVertex3f(-1,-1,1);
					
		glTexCoord2f(0.0,1.0);
		glVertex3f(-1,0.25,1);
					
		glTexCoord2f(1.0,1.0);
		glVertex3f(1,0.25,1);
				
		glTexCoord2f(1.0,0.0);
		glVertex3f(1,-1,1);
	glEnd();					
		
	/* right-back */
	glBegin(GL_QUADS);
		glTexCoord2f(0.0,0.0);
		glVertex3f(1,-1,-1);
					
		glTexCoord2f(0.0,1.0);
		glVertex3f(1,0.25,-1);
					
		glTexCoord2f(1.0,1.0);
		glVertex3f(1,0.25,1);
				
		glTexCoord2f(1.0,0.0);
		glVertex3f(1,-1,1);
	glEnd();
	
	/* left-back */
	glBegin(GL_QUADS);
		glTexCoord2f(0.0,0.0);
		glVertex3f(-1,-1,-1);
					
		glTexCoord2f(0.0,1.0);
		glVertex3f(-1,0.25,-1);
					
		glTexCoord2f(1.0,1.0);
		glVertex3f(1,0.25,-1);
				
		glTexCoord2f(1.0,0.0);
		glVertex3f(1,-1,-1);
	glEnd();
		
	/* left-front */
	glBegin(GL_QUADS);
		glTexCoord2f(0.0,0.0);
		glVertex3f(-1,-1,-1);
					
		glTexCoord2f(0.0,1.0);
		glVertex3f(-1,0.25,-1);
					
		glTexCoord2f(1.0,1.0);
		glVertex3f(-1,0.25,1);
				
		glTexCoord2f(1.0,0.0);
		glVertex3f(-1,-1,1);
	glEnd();
}

/*
=================
display
-renders output/scene to the screen
=================
*/
void display(void) 
{
	/* clear frame buffer and z-buffer */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	/* move to the center of the screen */
	glLoadIdentity();
	/* moves 0.5 units (up) on y-axis and -3.5 units (forward) on z-axis */
	glTranslatef(0.0, 0.5, -3.5);
	/* initial rotation corresponds to isometric drawing angle (30 degrees around x-axis) */
	glRotatef(30.0, 1.0, 0.0, 0.0);
	/* now rotate the camera upward (y-axis) at a 45 degree angle */
	glRotatef(45.0, 0.0, 1.0, 0.0);

	/* loads the texture(s) */
	textureLoader();
	/* draw the 3D box */
	boxCreator();
	
	/* make sure all previously issued commands are completed */
	glFlush(); 
	/* contents of back buffer are set to become that of front buffer (double buffering)*/
	glutSwapBuffers();
}

/*
=================
reshape
-This function ensures that our world are correctly resized, regardless of the size or 
 shape of the window.
-It works by clearing and redrawing the screen on each resize, with the glutReshapeFunc always updating the reshape function with the new width and height of the window.
=================
*/
void reshape(int w, int h)
{
	/* sets the viewport with the first two parameters the lower-left corner of the 
	    viewport rectangle, in pixels and the last two, the width and height of the 
	    viewport */
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);

	/* specifies the Projection matrix as the current matrix */
	glMatrixMode(GL_PROJECTION);
	/* move to the center of the screen */
	glLoadIdentity();

	/* sets up the  perspective projection matrix:
	   - the first parameter specifies the field of view (fov), an angle in degrees 
	     (in the y-direction)
	   - the second parameter is aspect ratio determining fov (in the x-direction)
	   - the third parameter is the distance from the viewer to the near clipping 
	     plane
	   - the last parameter is the distance from the viewer to the far clipping plane */
	gluPerspective(45, (GLdouble)w / (GLdouble)h, 0.05, 100);
	/* specifies the Modelview matrix as the current matrix */
	glMatrixMode(GL_MODELVIEW);
}

/*
=================
main
=================
*/
int main(int argc, char** argv)
{
	/* initialize the GLUT library. */
	glutInit(&argc, argv);
	/* sets the initial display mode: double buffering and a depth buffer. */
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	/* sets default window size */
	glutInitWindowSize(window_width, window_height);
	/* creates a top-level window with a caption */
	glutCreateWindow("...texture mapping...");
	/* reshape callback for the current window -> for use in reshape() */
	glutReshapeFunc(reshape);
	/* enables the OpenGL texturing feature */
	glEnable(GL_TEXTURE_2D);
	/* displays our contents in the window */
	glutDisplayFunc(display);
	/* enable hidden-surface removal in our case hiding the floor behind 
	   the wall facing the viewer - remove it to see the result */
	glEnable(GL_DEPTH_TEST);
	/* enters the GLUT event processing loop - will handle any callback */
	glutMainLoop();
	return 0;
}
//EOF